/**************************************************************************
*  This file is part of the TAL project (Tiny Abstraction Layer)
*
*  Copyright (c) 2016-2018 by Michael Fischer (www.emb4fun.de).
*  All rights reserved.
*
*  Redistribution and use in source and binary forms, with or without 
*  modification, are permitted provided that the following conditions 
*  are met:
*  
*  1. Redistributions of source code must retain the above copyright 
*     notice, this list of conditions and the following disclaimer.
*  2. Redistributions in binary form must reproduce the above copyright
*     notice, this list of conditions and the following disclaimer in the 
*     documentation and/or other materials provided with the distribution.
*  3. Neither the name of the author nor the names of its contributors may 
*     be used to endorse or promote products derived from this software 
*     without specific prior written permission.
*
*  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS 
*  "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT 
*  LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS 
*  FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL 
*  THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, 
*  INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, 
*  BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS 
*  OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED 
*  AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, 
*  OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF 
*  THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF 
*  SUCH DAMAGE.
*
***************************************************************************
*  History:
*
*  14.10.2016  mifi  First Version for the BeagleBone Black (B3).
**************************************************************************/
#if defined(USE_BOARD_B3)
#define __TALLED_C__

/*=======================================================================*/
/*  Include                                                              */
/*=======================================================================*/
#include <stdint.h>
#include <string.h>
#include "tal.h"

#include "soc_AM335x.h"
#include "hw_control_AM335x.h"
#include "hw_types.h"
#include "hw_cm_wkup.h"
#include "hw_cm_per.h"
#include "hw_gpio_v2.h"

/*=======================================================================*/
/*  All Structures and Common Constants                                  */
/*=======================================================================*/

/*
 * Control register
 */
#define CONTROL_CONF_MMODE_SHIFT       0
#define CONTROL_CONF_PUDEN_SHIFT       3
#define CONTROL_CONF_PUTYPESEL_SHIFT   4
#define CONTROL_CONF_RXACTIVE_SHIFT    5


/*
 * LED status
 */
#define LED_OFF   0
#define LED_ON    1

/*=======================================================================*/
/*  Definition of all local Data                                         */
/*=======================================================================*/

static uint8_t LedStatus[TAL_LED_CHANNEL_MAX];

/*=======================================================================*/
/*  Definition of all local Procedures                                   */
/*=======================================================================*/

/*************************************************************************/
/*  GPIO1ModuleClkConfig                                                 */
/*                                                                       */
/*  This function enables the functional clocks for GPIO1 instance.      */
/*  Based on the StarterWare example (platform\evmskAM335x\gpio.c).      */
/*                                                                       */
/*  In    : none                                                         */
/*  Out   : none                                                         */
/*  Return: none                                                         */
/*************************************************************************/
static void GPIO1ModuleClkConfig (void)
{
   /* Writing to MODULEMODE field of CM_PER_GPIO1_CLKCTRL register. */
   HWREG(SOC_CM_PER_REGS + CM_PER_GPIO1_CLKCTRL) |=
         CM_PER_GPIO1_CLKCTRL_MODULEMODE_ENABLE;

   /* Waiting for MODULEMODE field to reflect the written value. */
   while(CM_PER_GPIO1_CLKCTRL_MODULEMODE_ENABLE !=
         (HWREG(SOC_CM_PER_REGS + CM_PER_GPIO1_CLKCTRL) &
          CM_PER_GPIO1_CLKCTRL_MODULEMODE));    /*lint !e722*/
          
   /*
   ** Writing to OPTFCLKEN_GPIO_1_GDBCLK bit in CM_PER_GPIO1_CLKCTRL
   ** register.
   */
   HWREG(SOC_CM_PER_REGS + CM_PER_GPIO1_CLKCTRL) |=
         CM_PER_GPIO1_CLKCTRL_OPTFCLKEN_GPIO_1_GDBCLK;

   /*
   ** Waiting for OPTFCLKEN_GPIO_1_GDBCLK bit to reflect the desired
   ** value.
   */
   while(CM_PER_GPIO1_CLKCTRL_OPTFCLKEN_GPIO_1_GDBCLK !=
         (HWREG(SOC_CM_PER_REGS + CM_PER_GPIO1_CLKCTRL) &
          CM_PER_GPIO1_CLKCTRL_OPTFCLKEN_GPIO_1_GDBCLK));   /*lint !e722*/

   /*
   ** Waiting for IDLEST field in CM_PER_GPIO1_CLKCTRL register to attain the
   ** desired value.
   */
   while((CM_PER_GPIO1_CLKCTRL_IDLEST_FUNC <<
          CM_PER_GPIO1_CLKCTRL_IDLEST_SHIFT) !=
          (HWREG(SOC_CM_PER_REGS + CM_PER_GPIO1_CLKCTRL) &
           CM_PER_GPIO1_CLKCTRL_IDLEST));    /*lint !e722*/

   /*
   ** Waiting for CLKACTIVITY_GPIO_1_GDBCLK bit in CM_PER_L4LS_CLKSTCTRL
   ** register to attain desired value.
   */
   while(CM_PER_L4LS_CLKSTCTRL_CLKACTIVITY_GPIO_1_GDBCLK !=
         (HWREG(SOC_CM_PER_REGS + CM_PER_L4LS_CLKSTCTRL) &
          CM_PER_L4LS_CLKSTCTRL_CLKACTIVITY_GPIO_1_GDBCLK));   /*lint !e722*/
          
} /* GPIO1ModuleClkConfig */

/*=======================================================================*/
/*  All code exported                                                    */
/*=======================================================================*/

/*************************************************************************/
/*  tal_LEDInit                                                          */
/*                                                                       */
/*  Initialize the LEDs of the board.                                    */
/*                                                                       */
/*  In    : none                                                         */
/*  Out   : none                                                         */
/*  Return: none                                                         */
/*************************************************************************/
void tal_LEDInit (void)
{
   uint32_t Value;

   /* Clear status information */
   memset(LedStatus, LED_OFF, sizeof(LedStatus));
   
   /*
    * Initialize GPIO1 led's
    */
   
   /* Init GPIO clocks */
   GPIO1ModuleClkConfig();    

   
   /* Switch off all LEDs first to prevent glitches */   
   tal_LEDClear(TAL_LED_CHANNEL_1);
   tal_LEDClear(TAL_LED_CHANNEL_2);
   tal_LEDClear(TAL_LED_CHANNEL_3);
   tal_LEDClear(TAL_LED_CHANNEL_4);
   
   
   /* Setup Pad Control Registers */
   Value = (7 << CONTROL_CONF_MMODE_SHIFT)     |   /* Mode 7 */
           (1 << CONTROL_CONF_PUDEN_SHIFT)     |   /* Pull disabled */
           (1 << CONTROL_CONF_PUTYPESEL_SHIFT) |   /* Pullup */
           (0 << CONTROL_CONF_RXACTIVE_SHIFT);     /* Receiver disabled */ /*lint !e845*/
   
   HWREG(SOC_CONTROL_REGS + CONTROL_CONF_GPMC_A(5)) = Value;
   HWREG(SOC_CONTROL_REGS + CONTROL_CONF_GPMC_A(6)) = Value;
   HWREG(SOC_CONTROL_REGS + CONTROL_CONF_GPMC_A(7)) = Value;
   HWREG(SOC_CONTROL_REGS + CONTROL_CONF_GPMC_A(8)) = Value;
   
   /* Module is enabled, clocks are not gated. */
   HWREG(SOC_GPIO_1_REGS + GPIO_CTRL) &= ~GPIO_CTRL_DISABLEMODULE;

   /* Clear the corresponding bits in the GPIO_DATAOUT register, set outputs to low. */
   HWREG(SOC_GPIO_1_REGS + GPIO_CLEARDATAOUT) = ((1<<21) | (1<<22) | (1<<23) | (1<<24));
   
   /* The corresponding GPIO ports are configured as an output. */
   HWREG(SOC_GPIO_1_REGS + GPIO_OE) &= ~((1<<21) | (1<<22) | (1<<23) | (1<<24));

} /* tal_LEDInit */

/*************************************************************************/
/*  tal_LEDSet                                                           */
/*                                                                       */
/*  Set the given LED.                                                   */
/*                                                                       */
/*  In    : eChannel                                                     */
/*  Out   : none                                                         */
/*  Return: none                                                         */
/*************************************************************************/
void tal_LEDSet (TAL_LED_CHANNEL eChannel)
{
   /* Check for valid range */
   if ((uint8_t)eChannel < (uint8_t)TAL_LED_CHANNEL_MAX)
   {
      switch (eChannel)
      {
         case TAL_LED_CHANNEL_1: HWREG(SOC_GPIO_1_REGS + GPIO_SETDATAOUT) = (1<<21); break;
         case TAL_LED_CHANNEL_2: HWREG(SOC_GPIO_1_REGS + GPIO_SETDATAOUT) = (1<<22); break;
         case TAL_LED_CHANNEL_3: HWREG(SOC_GPIO_1_REGS + GPIO_SETDATAOUT) = (1<<23); break;
         case TAL_LED_CHANNEL_4: HWREG(SOC_GPIO_1_REGS + GPIO_SETDATAOUT) = (1<<24); break;
         default:
         {
            /* Do nothing */
            break;
         }
      }
      
      LedStatus[eChannel] = LED_ON;
   }
   
} /* hal_LEDSet */

/*************************************************************************/
/*  tal_LEDClear                                                         */
/*                                                                       */
/*  Clear the given LED.                                                 */
/*                                                                       */
/*  In    : eChannel                                                     */
/*  Out   : none                                                         */
/*  Return: none                                                         */
/*************************************************************************/
void tal_LEDClear (TAL_LED_CHANNEL eChannel)
{
   /* Check for valid range */
   if ((uint8_t)eChannel < (uint8_t)TAL_LED_CHANNEL_MAX)
   {
      switch (eChannel)
      {
         case TAL_LED_CHANNEL_1: HWREG(SOC_GPIO_1_REGS + GPIO_CLEARDATAOUT) = (1<<21); break;
         case TAL_LED_CHANNEL_2: HWREG(SOC_GPIO_1_REGS + GPIO_CLEARDATAOUT) = (1<<22); break;
         case TAL_LED_CHANNEL_3: HWREG(SOC_GPIO_1_REGS + GPIO_CLEARDATAOUT) = (1<<23); break;
         case TAL_LED_CHANNEL_4: HWREG(SOC_GPIO_1_REGS + GPIO_CLEARDATAOUT) = (1<<24); break;
         default:
         {
            /* Do nothing */
            break;
         }
      }
      
      LedStatus[eChannel] = LED_OFF;
   }
   
} /* tal_LEDClear */

/*************************************************************************/
/*  tal_LEDToggle                                                        */
/*                                                                       */
/*  Toggle the given LED.                                                */
/*                                                                       */
/*  In    : eChannel                                                     */
/*  Out   : none                                                         */
/*  Return: none                                                         */
/*************************************************************************/
void tal_LEDToggle (TAL_LED_CHANNEL eChannel)
{
   /* Check for valid range */
   if ((uint8_t)eChannel < (uint8_t)TAL_LED_CHANNEL_MAX)
   {
      if (LED_ON == LedStatus[eChannel])
      { 
         tal_LEDClear(eChannel); 
      }
      else
      {
         tal_LEDSet(eChannel); 
      }         
   }
   
} /* tal_LEDClear */

#endif /* USE_BOARD_B3 */

/*** EOF ***/
