/**************************************************************************
*  This file is part of the TCTS project (Tiny Cooperative Task Scheduler)
*
*  Copyright (c) 2014-2018 by Michael Fischer (www.emb4fun.de).
*  All rights reserved.
*
*  Redistribution and use in source and binary forms, with or without 
*  modification, are permitted provided that the following conditions 
*  are met:
*  
*  1. Redistributions of source code must retain the above copyright 
*     notice, this list of conditions and the following disclaimer.
*  2. Redistributions in binary form must reproduce the above copyright
*     notice, this list of conditions and the following disclaimer in the 
*     documentation and/or other materials provided with the distribution.
*  3. Neither the name of the author nor the names of its contributors may 
*     be used to endorse or promote products derived from this software 
*     without specific prior written permission.
*
*  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS 
*  "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT 
*  LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS 
*  FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL 
*  THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, 
*  INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, 
*  BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS 
*  OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED 
*  AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, 
*  OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF 
*  THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF 
*  SUCH DAMAGE.
*
***************************************************************************
*  History:
*
*  17.01.2014  mifi  First Version.
*  15.02.2015  mifi  Expanded for Cortex-M.
*  27.09.2015  mifi  Corrected AddTaskToWaitList. TCBFifoAddEnd must used
*                    instead of TCBFifoAddPrio for correct task order.
*  14.11.2015  mifi  Expanded for Cortex-A8.
*  15.11.2015  mifi  Expanded for Nios II.
*  04.01.2016  mifi  Added more timezone functionality.
*                    Corrected "WaitList" bug.
*  10.01.2016  mifi  Rework semaphore handling.
*                    Rework OSTaskCreate, schedule new task if
*                    scheduler is not locked.
*                    Added OSTaskGetPriority and OSTimeDlyUntil.
*                    Added event functionality. 
*  04.02.2016  mifi  Remove OSSemaBroadcast and OSSemaBroadcastAsync.
*                    Change OSSemaCreate parameter from uint32 to int32.
*  08.02.2016  mifi  Added "Performance optimisation" switch.
*  12.02.2016  mifi  Added the some more functionality:
*                    - OSTaskTerminateRequest and OSTaskShouldTerminate
*                    - OSTimerSet 
*                    - OSSemaReset
*  05.05.2016  mifi  Added incomplete Mailbox functionality, only:
*                    - OSMboxCreate
*                    - OSMboxDelete
*                    - OSMboxPostFromInt
*                    - OSMboxWait
*  16.06.2016  mifi  Added Cortex-M0+ support.
*  12.05.2018  mifi  Added TaskScheduleExit support.
*  23.06.2018  mifi  Change version pattern.
*                    Rework OS_WAIT_INFINITE handling.
*                    Added mutex functionality
*                    - OSMutexCreate
*                    - OSMutexSignal
*                    - OSMutexSignalFromInt
*                    - OSMutexWait
*                    Change version to v0.18.1
**************************************************************************/
#define __TCTS_C__

/*=======================================================================*/
/*  Include                                                              */
/*=======================================================================*/
#include <string.h>
#include <stdint.h>

#include "tcts.h"
#include "tal.h"

/*=======================================================================*/
/*  All extern data                                                      */
/*=======================================================================*/

/*=======================================================================*/
/*  All Structures and Common Constants                                  */
/*=======================================================================*/

/*
 * Check if "performacne optimisation" is enabled.
 * If this option is enabled the statistic functionality is limited.
 * No task based time information are available in this case.
 */
#if !defined(TCTS_ENABLE_PERFORMANCE_OPTIMISATION_NO_STATISTIC)
   Error: TCTS_ENABLE_PERFORMANCE_OPTIMISATION_NO_STATISTIC must be defined in tcts_conf.h;
#else
#define _PERF_OPT_NO_STATISTIC   TCTS_ENABLE_PERFORMANCE_OPTIMISATION_NO_STATISTIC
#endif    

/*
 * Idle and statistic stack size
 */
#if !defined(OS_IDLE_STACK_SIZE)
#define OS_IDLE_STACK_SIZE 512
#endif

#if !defined(OS_STAT_STACK_SIZE)
#define OS_STAT_STACK_SIZE 512
#endif

/*
 * Idle and statistic task priority
 */
#define IDLE_TASK_PRIO     255
#define STAT_TASK_PRIO     2

/*
 * Test if the fifo is empty
 */
#define IsFifoEmpty(_f)       (((_f)->pIn == NULL) ? 1 : 0)

/*
 * Set state of a task
 */
#if (_PERF_OPT_NO_STATISTIC >= 1) 
#define SET_TASK_STATE(_a,_b)
#else
#define SET_TASK_STATE(_a,_b) { _a->State = _b; }
#endif

/*=======================================================================*/
/*  Definition of all global Data                                        */
/*=======================================================================*/   

/*=======================================================================*/
/*  Definition of all local Data                                         */
/*=======================================================================*/

/*
 * Ide and Statistic task variables
 */
static OS_STACK(IdleStack, OS_IDLE_STACK_SIZE);
static OS_STACK(StatStack, OS_STAT_STACK_SIZE);

static os_tcb_t TCBIdle;
static os_tcb_t TCBStat;

/*
 * Scheduler lock status
 */
static uint8_t bIsSchedLocked;

/*
 * Time variables
 */
static volatile uint32_t dSystemTick = 0;
static volatile uint32_t dSecTick    = 0;
static volatile uint32_t dUnixtime   = 0;

static int16_t   wTimezoneID           = 0;
static int16_t   wTimezoneOffsetMin    = 0;
static int32_t   dTimezoneOffsetSec    = 0;
static int32_t   dTimezoneDstOffsetSec = 0;

/*
 * Some statistics variables
 */
static uint8_t   bStatEnabled       = 0; 
static uint8_t   bStatCPULoad       = 0;
static uint32_t  dStatTotalTaskTime = 0;
static uint32_t  dStatHirResPeriod  = 0;

/* 
 * The TaskList contains all the tasks of the system. 
 */
static os_tcb_t *pTaskList = NULL;

/* 
 * The ReadyList contains the tasks which are ready to run. 
 */
static os_tcb_fifo_t ReadyList;

/*
 * The WaitList contains the tasks which are waiting
 * with timeout, e.g. OSTimeDly, OSSemaWait or OSEventWait.
 */
static os_tcb_fifo_t WaitList;

/*
 * This is the actual running task.
 */
static os_tcb_t *RunningTask = NULL;

/*
 * This is the task which will run after the context switch.
 */
static os_tcb_t *NewTask = NULL;

static OS_TIMER_FUNC UserTimer = NULL;
static uint32_t      UserTimerTimeout = 0;

/*=======================================================================*/
/*  Definition of all local Procedures                                   */
/*=======================================================================*/

/*
 * Check for ARM7 style CPU
 */
#if defined(__ARM_ARCH_4T__) || defined(__ARM_ARCH_7A__)
#include "tcts_arm.c"
#endif

/*
 * Check for Cortex-M style CPU
 */
#if defined(__ARM_ARCH_7EM__) || defined(__ARM_ARCH_7M__)
#include "tcts_cm.c"
#endif

/*
 * Check for Cortex-M0 style CPU
 */
#if defined(__ARM_ARCH_6M__)
#include "tcts_cm0plus.c"
#endif

/*
 * Check for Nios II style CPU
 */
#if defined(__NIOS__)
#include "tcts_nios.c"
#endif

/*************************************************************************/
/*  GetStackFreeCount                                                    */
/*                                                                       */
/*  The stack was initialized with 0xCC.                                 */
/*  Unused entries contains still 0xCC.                                  */
/*                                                                       */
/*  In    : pStart, pEnd                                                 */
/*  Out   : none                                                         */
/*  Return: Free entries of the stack                                    */
/*************************************************************************/
static uint32_t GetStackFreeCount (uint8_t *pStart, uint8_t *pEnd)
{
   uint32_t dFreeCount = 0;
   
   while (pStart < pEnd)
   {
      if (0xCC == *pStart)
      {
         dFreeCount++;   
      }
      else
      {
         break;
      }
      pStart++;
   }
   
   return(dFreeCount);
} /* GetStackFreeCount */

/*************************************************************************/
/*  TCBFifoAddPrio                                                       */
/*                                                                       */
/*  Add the task to the FIFO, with respect to the priority.              */ 
/*                                                                       */
/*  Note: Interrupts are disabled and must be disabled.                  */
/*                                                                       */
/*  In    : pFifo, pTask                                                 */
/*  Out   : none                                                         */
/*  Return: none                                                         */
/*************************************************************************/
static __inline__ void TCBFifoAddPrio (os_tcb_fifo_t *pFifo, os_tcb_t *pTask)
{
   os_tcb_t *pList;
   int        TaskWasAdded = 0;

#if 0 /* TODO: double check if it is a good idea */
   /* The Idle task must not added to any lists */
   if (&TCBIdle == pTask)
   {
      return;
   }
#endif   

   /* Check if the fifo is empty */
   if (NULL == pFifo->pIn)
   {
      /* Fifo is empty */
      pTask->pPrev = NULL;    /* Task has no prev */
      pTask->pNext = NULL;    /* and no next */
      
      pFifo->pIn  = pTask;    /* Fifo in and out */      
      pFifo->pOut = pTask;    /* points to the new task */
   }
   else
   {
      pList = pFifo->pIn;
      while (pList != NULL)
      {
         if (pTask->nPrio >= pList->nPrio)
         {
            if (pList->pPrev == NULL)
            {
               /* Insert at the front of the fifo */
               pTask->pPrev = NULL;          /* This is the new first task, therefore no prev. */
               pTask->pNext = pFifo->pIn;    /* Link the new task in front of the fifo. */
               pFifo->pIn->pPrev = pTask;    /* Link the second task (old first) back to the new first task. */
               pFifo->pIn = pTask;           /* Set the fifo->pIn to the new first task. */
            }
            else
            {
               /* Insert in the middle of the fifo */
               pTask->pNext = pList;
               pTask->pPrev = pList->pPrev;
               
               pList->pPrev->pNext = pTask;
               pList->pPrev        = pTask;
            }
            TaskWasAdded = 1;
            break;
         } /* end if (pTask->Prio >= pList->Prio) */
      
         pList = pList->pNext;
      } /* end while (pList != NULL) */
      
      /* Check if task was not added */
      if (0 == TaskWasAdded)
      {
         /* Add Task at the end of the fifo */
         pTask->pNext = NULL;
         pTask->pPrev = pFifo->pOut;
         
         pFifo->pOut->pNext = pTask;
         pFifo->pOut        = pTask;
      }
   }
   
} /* TCBFifoAddPrio */

/*************************************************************************/
/*  TCBFifoRemove                                                        */
/*                                                                       */
/*  Remove a task from the FIFO, output side, and return the task.       */
/*                                                                       */
/*  Note: Interrupts are disabled and must be disabled.                  */
/*                                                                       */
/*  In    : pFifo                                                        */
/*  Out   : none                                                         */
/*  Return: pTask                                                        */
/*************************************************************************/
static __inline__ os_tcb_t *TCBFifoRemove (os_tcb_fifo_t *pFifo)
{
   os_tcb_t *pTask;

   /* Check if the fifo is empty */
   if (NULL == pFifo->pOut)
   {
      /* Error */
      TAL_FAILED();
      return(NULL);
   }
   
   /* Check if only one element is in the fifo */
   if (pFifo->pIn == pFifo->pOut)
   {
      /* Only one element is in the fifo */
      pTask = pFifo->pOut;

      /* Clear the fifo in/out */
      pFifo->pIn  = NULL;
      pFifo->pOut = NULL;
   }
   else
   {
      pTask = pFifo->pOut;          /* Get the last task of the fifo */
      pFifo->pOut = pTask->pPrev;   /* Move the last pointer to the task before the end */
      pFifo->pOut->pNext = NULL;    /* Remove the link from the new last element to the next one */
   }

#if 1 /* TODO: Is this realy needed ? */
   /* Clear fifo pointer */
   pTask->pPrev = NULL;
   pTask->pNext = NULL;
#endif   
         
   return(pTask);
} /* TCBFifoRemove */

/*************************************************************************/
/*  TCBFifoRemoveMiddle                                                  */
/*                                                                       */
/*  Remove a task from the FIFO.                                         */
/*                                                                       */
/*  Note: Interrupts are disabled and must be disabled.                  */
/*                                                                       */
/*  In    : pFifo, pTask                                                 */
/*  Out   : none                                                         */
/*  Return: none                                                         */
/*************************************************************************/
static __inline__ void TCBFifoRemoveMiddle (os_tcb_fifo_t *pFifo, os_tcb_t *pTask)
{
   /* Check if fifo is empty */
   if (NULL == pFifo->pIn)
   {
      /* Error */
      TAL_FAILED();
   }

   /* Check if only one element is in the list */
   if (pFifo->pIn == pFifo->pOut)
   {
      /* Clear the fifo in/out */
      pFifo->pIn  = NULL;
      pFifo->pOut = NULL;
   }
   else
   {
      /* Check if the task is in the middle of the list */
      if      ((pTask->pPrev != NULL) && (pTask->pNext != NULL))
      {
         pTask->pNext->pPrev = pTask->pPrev;
         pTask->pPrev->pNext = pTask->pNext;
      }
      /* Check for the first element */
      else if (pTask->pPrev == NULL)
      {
         pTask->pNext->pPrev = NULL;
         pFifo->pIn = pTask->pNext;
      }
      /* Check for the last element */
      else if (pTask->pNext == NULL)
      {
         pTask->pPrev->pNext = NULL;
         pFifo->pOut = pTask->pPrev;
      }
   }

   pTask->pNext = NULL;
   pTask->pPrev = NULL;
} /* TCBFifoRemoveMiddle */

/*************************************************************************/
/*  TaskListAdd                                                          */
/*                                                                       */
/*  Add the task to the TaskList.                                        */ 
/*                                                                       */
/*  Note: Interrupts are disabled and must be disabled.                  */
/*                                                                       */
/*  In    : pTask                                                        */
/*  Out   : none                                                         */
/*  Return: none                                                         */
/*************************************************************************/
static void TaskListAdd (os_tcb_t *pTask)
{
   /* Check if the list is empty */
   if (NULL == pTaskList)
   {
      /* List is empty */
      pTask->pTaskPrev = NULL;
      pTask->pTaskNext = NULL;

      pTaskList = pTask;
   }
   else
   {
      pTask->pTaskPrev = NULL;
      
      /*
       * Link the new task in front of the fifo.
       */
      pTask->pTaskNext = pTaskList;
      
      /* 
       * Link the old "first task" back to the new task.
       */
      pTaskList->pTaskPrev = pTask;
   
      /*
       * Set the new fifo->pIn.
       */
      pTaskList = pTask;
   }
   
} /* TaskListAdd */

/*************************************************************************/
/*  TaskListRemove                                                       */
/*                                                                       */
/*  Remove a task from the TaskList.                                     */
/*                                                                       */
/*  Note: Interrupts are disabled and must be disabled.                  */
/*                                                                       */
/*  In    : pTask                                                        */
/*  Out   : none                                                         */
/*  Return: none                                                         */
/*************************************************************************/
static void TaskListRemove (os_tcb_t *pTask)
{
   /* Check if only one element is in the list */
   if ((NULL == pTaskList->pTaskPrev) && (NULL == pTaskList->pTaskNext))
   {
      /* Clear the list */
      pTaskList = NULL;
   }
   else
   {
      /* Check if the task is in the middle of the list */
      if      ((pTask->pTaskPrev != NULL) && (pTask->pTaskNext != NULL))
      {
         pTask->pTaskNext->pTaskPrev = pTask->pTaskPrev;
         pTask->pTaskPrev->pTaskNext = pTask->pTaskNext;
      }
      /* Check for the first element */
      else if (pTask->pTaskPrev == NULL)
      {
         pTask->pTaskNext->pTaskPrev = NULL;
         pTaskList = pTask->pTaskNext;
      }
      /* Check for the last element */
      else if (pTask->pTaskNext == NULL)
      {
         pTask->pTaskPrev->pTaskNext = NULL;
      }
   }
   
} /* TaskListRemove */

/*************************************************************************/
/*  AddTaskToWaitList                                                    */
/*                                                                       */
/*  Add the task to the WaitList fifo.                                   */
/*                                                                       */
/*  Note: Interrupts are disabled and must be disabled.                  */
/*                                                                       */
/*  In    : pTask                                                        */
/*  Out   : none                                                         */
/*  Return: none                                                         */
/*************************************************************************/
static __inline__ void AddTaskToWaitList (os_tcb_t *pTask)
{
   os_tcb_fifo_t *pFifo = &WaitList;
    
   SET_TASK_STATE(pTask, OS_TASK_STATE_WAITING);
   
   /*
    * For the correct order of task processing the task must be 
    * added at the end. For example, if there exist three tasks, A, B and C 
    * with the same priority. And the tasks was started in the ABC order the
    * WaitList looks like A->B->C after all tasks was added to the WaitList
    * now.
    *
    * CheckWaitList will move the tasks later from the WaitList to  
    * the ReadyList if the timeout is expired. But this will be
    * done with TCBFifoAddPrio. The ReadyList will look like C->B->A.
    *
    * The task which is started is taken by the scheduler from the 
    * end of the ReadyList. Now the order of task processing is correct.
    */

   /* The Idle task must not added to any lists */
   if (&TCBIdle == pTask)
   {
      return;
   }

   /* Check if the fifo is empty */
   if (NULL == pFifo->pIn)
   {
      /* Fifo is empty */
      pTask->pWaitPrev = NULL;         /* Task has no prev */
      pTask->pWaitNext = NULL;         /* and no next */
      
      pFifo->pIn  = pTask;             /* Fifo in and out */      
      pFifo->pOut = pTask;             /* points to the new task */
   }
   else   
   {
      /* Fifo is not empty */
      pTask->pWaitNext = NULL;         /* This is the new last task, therefore no next. */    
      pTask->pWaitPrev = pFifo->pOut;  /* Link the new task at the end of the fifo. */
      pFifo->pOut->pWaitNext = pTask;  /* Link the old last to the new last. */
      pFifo->pOut = pTask;             /* Set the fifo-pOut to the new last task. */
   }
   
} /* AddTaskToWaitList */

/*************************************************************************/
/*  WaitListRemove                                                       */
/*                                                                       */
/*  Remove a task from the WaitList fifo.                                */
/*                                                                       */
/*  Note: Interrupts are disabled and must be disabled.                  */
/*                                                                       */
/*  In    : pTask                                                        */
/*  Out   : none                                                         */
/*  Return: none                                                         */
/*************************************************************************/
static void WaitListRemove (os_tcb_t *pTask)
{
   os_tcb_fifo_t *pFifo = &WaitList;
   
   /* Check if fifo is empty */
   if (NULL == pFifo->pIn)
   {
      /* Error */
      TAL_FAILED();
   }
   
   /* Check if only one element is in the list */
   if (pFifo->pIn == pFifo->pOut)
   {
      /* Clear the fifo in/out */
      pFifo->pIn  = NULL;
      pFifo->pOut = NULL;
   }
   else
   {
      /* Check if the task is in the middle of the list */
      if      ((pTask->pWaitPrev != NULL) && (pTask->pWaitNext != NULL))
      {
         pTask->pWaitNext->pWaitPrev = pTask->pWaitPrev;
         pTask->pWaitPrev->pWaitNext = pTask->pWaitNext;
      }
      /* Check for the first element */
      else if (pTask->pWaitPrev == NULL)
      {
         pTask->pWaitNext->pWaitPrev = NULL;
         pFifo->pIn = pTask->pWaitNext;
      }
      /* Check for the last element */
      else if (pTask->pWaitNext == NULL)
      {
         pTask->pWaitPrev->pWaitNext = NULL;
         pFifo->pOut = pTask->pWaitPrev;
      }
   }

   pTask->pWaitPrev = NULL;
   pTask->pWaitNext = NULL;   
} /* WaitListRemove */

/*************************************************************************/
/*  AddTaskToReadyList                                                   */
/*                                                                       */
/*  Add the task to the Ready list                                       */
/*                                                                       */
/*  Note: Interrupts are disabled and must be disabled.                  */
/*                                                                       */
/*  In    : pTask                                                        */
/*  Out   : none                                                         */
/*  Return: none                                                         */
/*************************************************************************/
static __inline__ void AddTaskToReadyList (os_tcb_t *pTask)
{
   SET_TASK_STATE(pTask, OS_TASK_STATE_READY);
   TCBFifoAddPrio(&ReadyList, pTask);   
} /* AddTaskToReadyList */

/*************************************************************************/
/*  GetReadyTask                                                         */
/*                                                                       */
/*  Get the task which is ready.                                         */
/*                                                                       */
/*  Note: Interrupts are disabled and must be disabled.                  */
/*                                                                       */
/*  In    : pTask                                                        */
/*  Out   : none                                                         */
/*  Return: none                                                         */
/*************************************************************************/
static os_tcb_t *GetReadyTask (void)
{
   os_tcb_t *pTask;

#if 0 /* TODO: double check if it is a good idea */
   /* In case the ReadyList fifo is empty, return the Idle task */
   if (NULL == ReadyList.pIn)
   {
      pTask = &TCBIdle;
   }
   else
#endif   
   {
      pTask = TCBFifoRemove(&ReadyList);
   }

   return(pTask);
} /* GetReadyTask */

/*************************************************************************/
/*  HandleWaitList                                                       */
/*                                                                       */
/*  The WaitList is used for handling a timeout or a delay.              */
/*                                                                       */
/*  Go through the WaitList and decrease the WaitCount. Check if the     */
/*  task from the WaitList can be moved to the ReadyList.                */
/*                                                                       */
/*  Note: Interrupts are disabled and must be disabled.                  */
/*                                                                       */
/*  In    : none                                                         */
/*  Out   : none                                                         */
/*  Return: none                                                         */
/*************************************************************************/
static void HandleWaitList (void)
{
   os_tcb_t *pTask = WaitList.pIn;
   os_tcb_t *pNext;

   while (pTask != NULL)
   {
      /* Decrease the WaitCount if allowed */
      if ((pTask->dTimeoutTicks > 0) && (pTask->dTimeoutTicks != OS_WAIT_INFINITE))
      {
         pTask->dTimeoutTicks--;
      }
   
      /* Check if the timeout is expired */
      if (0 == pTask->dTimeoutTicks)
      {
         /* 
          * Save the next pointer of the task.
          * pTask->pNext could not be used later, because it wil be changed
          * by the TCBFifoAdd before we can use the Next for the loop.
          */
         pNext = pTask->pWaitNext;
         
         /* Remove the task from the WaitList */
         WaitListRemove(pTask);
         
         /* Check if the task was waiting for a semaphore */
         if (pTask->pSemaWait)
         {
            /* 
             * This is a timeout of the semaphore, 
             * the task must be removed from the semaphore list too.
             */
            TCBFifoRemoveMiddle(&pTask->pSemaWait->Fifo, pTask);
            pTask->pSemaWait   = NULL;
            pTask->nReturnCode = OS_RC_TIMEOUT;  /* Return code for OSSemaWait */
         }

         /* Check if the task was waiting for a mutex */
         if (pTask->pMutexWait)
         {
            /* 
             * This is a timeout of the mutex, 
             * the task must be removed from the mutex list too.
             */
            TCBFifoRemoveMiddle(&pTask->pMutexWait->Fifo, pTask);
            pTask->pMutexWait  = NULL;
            pTask->nReturnCode = OS_RC_TIMEOUT;  /* Return code for OSMutexWait */
         }

         /* Check if the task was waiting for an event */
         if (pTask->pEventWait)
         {
            /* 
             * This is a timeout of the event, 
             * the task must be removed from the event list too.
             */
            TCBFifoRemoveMiddle(&pTask->pEventWait->Fifo, pTask);
            pTask->pEventWait  = NULL;
            pTask->nReturnCode = OS_RC_TIMEOUT;  /* Return code for OSEventWait */
         }
         
         /* Add the task to the ReadyList */            
         AddTaskToReadyList(pTask);
         
         /* Loop through the list */
         pTask = pNext;
      }
      else
      {
         /* Loop through the list */
         pTask = pTask->pWaitNext;
      }
   } /* end while (pTask != NULL) */
   
} /* HandleWaitList */

/*************************************************************************/
/*  StatUpdateTime                                                       */
/*                                                                       */
/*  Update the statistic time information from a task.                   */
/*                                                                       */
/*  Note: Interrupts are disabled and must be disabled.                  */
/*                                                                       */
/*  In    : pTask                                                        */
/*  Out   : none                                                         */
/*  Return: none                                                         */
/*************************************************************************/
static void StatUpdateTime (os_tcb_t *pTask)
{
   uint16_t wStartTick;
   uint16_t wEndTick;
   uint16_t wStartTimer;
   uint16_t wEndTimer;
   uint16_t wDeltaTick;
   uint16_t wDeltaTimer;
   
   wStartTimer = pTask->dStatStartTime & 0xFFFF;
   wStartTick  = pTask->dStatStartTime >> 16;
   wEndTimer   = pTask->dStatEndTime & 0xFFFF;
   wEndTick    = pTask->dStatEndTime >> 16;
   
   wDeltaTick = wEndTick - wStartTick;

   if (wEndTimer > wStartTimer)
   {
      wDeltaTimer = wEndTimer - wStartTimer;
   }
   else
   {
      wDeltaTimer = (uint16_t)(dStatHirResPeriod - wStartTimer) + wEndTimer;
   }
   
   pTask->dStatTotalTime += (wDeltaTick * dStatHirResPeriod) + wDeltaTimer;
       
} /* StatUpdateTime */

/*************************************************************************/
/*  TaskSchedule                                                         */
/*                                                                       */
/*  Must be called with interrupts disabled.                             */
/*                                                                       */
/*  Note: Interrupts are disabled and must be disabled.                  */
/*                                                                       */
/*  In    : none                                                         */
/*  Out   : none                                                         */
/*  Return: none                                                         */
/*************************************************************************/
static void TaskSchedule (void)
{ 
   /*
    * Get new task from the ReadyList.
    */
   NewTask = GetReadyTask();
   
   /*
    * If NewTask == RunningTask, we does not need a task switch
    */
   if (NewTask != RunningTask)
   {
#if (_PERF_OPT_NO_STATISTIC >= 1)
      /* Context switch needed */
      ContextSwitch();
#else   
      if (!bStatEnabled)
      {
         /* Context switch needed */
         ContextSwitch();
      }
      else
      {
         /* Get task end time for statistics */
         RunningTask->dStatEndTime = tal_CPUStatGetHiResCnt();
         StatUpdateTime(RunningTask);
   
         /* Context switch needed */
         ContextSwitch();    

         /* Get task start time for statistics */
         RunningTask->dStatStartTime = tal_CPUStatGetHiResCnt();
      }
#endif /* (_PERF_OPT_NO_STATISTIC >= 1) */      
   }
   
} /* TaskSchedule */

/*************************************************************************/
/*  TaskScheduleExit                                                     */
/*                                                                       */
/*  Must be called with interrupts disabled.                             */
/*                                                                       */
/*  Note: Interrupts are disabled and must be disabled.                  */
/*                                                                       */
/*  In    : none                                                         */
/*  Out   : none                                                         */
/*  Return: none                                                         */
/*************************************************************************/
static void TaskScheduleExit (void)
{ 
   /*
    * Get new task from the ReadyList.
    */
   NewTask = GetReadyTask();
   
   /*
    * If NewTask == RunningTask, we does not need a task switch
    */
   if (NewTask != RunningTask)
   {
#if (_PERF_OPT_NO_STATISTIC >= 1)
      /* Context switch needed */
      ContextSwitchExit();
#else   
      if (!bStatEnabled)
      {
         /* Context switch needed */
         ContextSwitchExit();
      }
      else
      {
         /* Get task end time for statistics */
         RunningTask->dStatEndTime = tal_CPUStatGetHiResCnt();
         StatUpdateTime(RunningTask);
   
         /* Context switch needed */
         ContextSwitchExit();    

         /* Get task start time for statistics */
         RunningTask->dStatStartTime = tal_CPUStatGetHiResCnt();
      }
#endif /* (_PERF_OPT_NO_STATISTIC >= 1) */      
   }
   
} /* TaskScheduleExit */

/*************************************************************************/
/*  StatTask                                                             */
/*                                                                       */
/*  In    : Task parameter                                               */
/*  Out   : none                                                         */
/*  Return: never                                                        */
/*************************************************************************/
static void StatTask (void *pParam)
{   
   os_tcb_t *pTask;
   uint32_t  dIdleTime = 0;
   
   (void)pParam;
   
   bStatEnabled      = 1;
   dStatHirResPeriod = tal_CPUStatGetHiResPeriod();
   
   while (1)
   {
      if (bStatEnabled)
      {
         /* Get total task time */
         dStatTotalTaskTime = 0;
      
         pTask = pTaskList;
         while (pTask != NULL)
         {
            /* Count total TaskTime */
            dStatTotalTaskTime += pTask->dStatTotalTime;
         
            /* Find Idle time */
            if (IDLE_TASK_PRIO == pTask->nPrio)
            {
               dIdleTime = pTask->dStatTotalTime;
            }
      
            pTask->dStatTotalLastTime = pTask->dStatTotalTime;
            pTask->dStatTotalTime     = 0;
         
            pTask = pTask->pTaskNext;
         }
         
         /*
          * Calculate CPU load like:
          * 
          *   bStatCPULoad = 100 - (uint8_t)((dIdleTime * 100) / dStatTotalTaskTime)
          *
          * But this makes problems with big counts. Therefore 
          * dIdleTime will not multiplied instead dStatTotalTaskTime
          * will be divided by 100.
          *
          *   bStatCPULoad = 100 - (uint8_t)(dIdleTime / (dStatTotalTaskTime / 100))
          */
         bStatCPULoad = 100 - (uint8_t)(dIdleTime / (dStatTotalTaskTime / 100));
      }         
      
      OSTimeDly(1000);
   }
   
} /* StatTask */

/*************************************************************************/
/*  IdleTask                                                             */
/*                                                                       */
/*  In    : Task parameter                                               */
/*  Out   : none                                                         */
/*  Return: never                                                        */
/*************************************************************************/
static void IdleTask (void *pParam)
{   
   (void)pParam;
   
   while (1)
   {
      /* Do not call any other OS function than OSTaskYield here */
      OSTaskYield();
   }
   
} /* IdleTask */

/*=======================================================================*/
/*  All code exported                                                    */
/*=======================================================================*/

/*************************************************************************/
/*  OSInit                                                               */
/*                                                                       */
/*  Init the "Cooperative Task Scheduler".                               */
/*                                                                       */
/*  In    : none                                                         */
/*  Out   : none                                                         */
/*  Return: none                                                         */
/*************************************************************************/
void OSInit (void)
{
   /* Init the Ready and Wait List */
   memset(&ReadyList, 0x00, sizeof(ReadyList));
   memset(&WaitList,  0x00, sizeof(WaitList));
   
   bIsSchedLocked = 1;     /* Used in tcts_cm, _arm and _nios */

   /* Create the Idle task */
   OSTaskCreate(&TCBIdle, IdleTask, NULL, IDLE_TASK_PRIO,
                IdleStack, sizeof(IdleStack), "[idle]");

   /* 
    * Set dStatTotalTaskTime to 1, to prevent
    * division by 0 in case statistic is not enabled.
    */
   dStatTotalTaskTime = 1;                
                
   (void)bIsSchedLocked;   /* Prevent lint warning */
} /* OSInit */

/*************************************************************************/
/*  OSSysTickStart                                                       */
/*                                                                       */
/*  Start the SysTick.                                                   */
/*                                                                       */
/*  In    : none                                                         */
/*  Out   : none                                                         */
/*  Return: never                                                        */
/*************************************************************************/
void OSSysTickStart (void)
{
   tal_CPUSysTickStart();
} /* OSSysTickStart */

/*************************************************************************/
/*  OSOutputRuntimeStackInfo                                             */
/*                                                                       */
/*  Output the runtime stack information.                                */
/*                                                                       */
/*  In    : none                                                         */
/*  Out   : none                                                         */
/*  Return: none                                                         */
/*************************************************************************/
void OSOutputRuntimeStackInfo (void)
{
   OutputRuntimeStackInfo();   
} /* OSOutputRuntimeStackInfo */

/*************************************************************************/
/*  OSOutputTaskInfo                                                     */
/*                                                                       */
/*  Output the Stack and Time information from the tasks.                */
/*                                                                       */
/*  In    : none                                                         */
/*  Out   : none                                                         */
/*  Return: none                                                         */
/*************************************************************************/
void OSOutputTaskInfo (void)
{
   os_tcb_t *pTCB;
   uint8_t  *pStart;
   uint8_t  *pEnd;
   uint32_t  dSize;
   uint32_t  dFree;
   uint32_t  dTotalTime;
   uint16_t  wCPUTimePercent;
   uint16_t  wTime1 = 0;
   uint16_t  wTime2 = 0;
   
   TAL_PRINTF("*** Task Info ***\n");
   TAL_PRINTF("\n");
   TAL_PRINTF("Task              Prio   Size   Used   Free   CPU %%\n");
   TAL_PRINTF("====================================================\n");

   /* Display "IRQ" Stack */
   pStart = GetIRQStackStart();
   pEnd   = GetIRQStackEnd();
   dSize  = (uint32_t)pEnd - (uint32_t)pStart;
   dFree  = GetStackFreeCount(pStart, pEnd);
   
   TAL_PRINTF("%-16s  ", "- IRQ -");
   TAL_PRINTF("  --   %4d   %4d   %4d   ----\n", dSize, (dSize - dFree), dFree);


   /* Get start of list */
   pTCB = OSTaskGetList(); 
   
   dTotalTime = OSStatGetTotalTaskTime();

   while (pTCB != NULL)
   {
      /* Do not handle Statistic and Idle here, will be handled at the end */
      if ((pTCB != &TCBStat) && (pTCB != &TCBIdle))
      {
         if (pTCB->Name[0] != 0)
         {
            TAL_PRINTF("%-16s  ", pTCB->Name);
         }
         else
         {
            TAL_PRINTF("%-16s  ", "----------------");
         }

         dFree = OSGetStackInfo (pTCB, &dSize);
      
         wCPUTimePercent = (uint16_t)(((uint64_t)pTCB->dStatTotalLastTime * 1000) / dTotalTime);
         wTime1 = wCPUTimePercent / 10;
         wTime2 = wCPUTimePercent % 10;
      
         TAL_PRINTF("%4d   %4d   %4d   %4d   %2d.%01d\n", pTCB->nPrio, dSize, (dSize - dFree), dFree, wTime1, wTime2);
      }         
      
      pTCB = pTCB->pTaskNext;
   }   
   
   /* 
    * Display Statistic task info
    */
   if (bStatEnabled)
   {
      pTCB  = &TCBStat;
      dFree = OSGetStackInfo (&TCBStat, &dSize);
      
      wCPUTimePercent = (uint16_t)(((uint64_t)pTCB->dStatTotalLastTime * 1000) / dTotalTime);
      wTime1 = wCPUTimePercent / 10;
      wTime2 = wCPUTimePercent % 10;
      
      TAL_PRINTF("%-16s  ", pTCB->Name);
      TAL_PRINTF("%4d   %4d   %4d   %4d   %2d.%01d\n", pTCB->nPrio, dSize, (dSize - dFree), dFree, wTime1, wTime2);
   }      

   /* 
    * Display Idle task info
    */
   pTCB  = &TCBIdle;
   dFree = OSGetStackInfo (&TCBIdle, &dSize);
   
   wCPUTimePercent = (uint16_t)(((uint64_t)pTCB->dStatTotalLastTime * 1000) / dTotalTime);
   wTime1 = wCPUTimePercent / 10;
   
   TAL_PRINTF("%-16s  ", pTCB->Name);
   TAL_PRINTF("%4d   %4d   %4d   %4d   %2d.%01d\n", pTCB->nPrio, dSize, (dSize - dFree), dFree, wTime1, wTime2);
         
   TAL_PRINTF("\n--- CPU load: %d%% ---\n", OSStatGetCPULoad());
         
   (void)dSize;
   (void)dFree;
   (void)wTime1;
   (void)wTime2;
         
   TAL_PRINTF("\n");   
} /* OSOutputTaskInfo */

/*************************************************************************/
/*  OSInit                                                               */
/*                                                                       */
/*  Init the "Cooperative Task Scheduler".                               */
/*                                                                       */
/*  In    : none                                                         */
/*  Out   : none                                                         */
/*  Return: none                                                         */
/*************************************************************************/
uint32_t OSGetStackInfo (os_tcb_t *pTCB, uint32_t *pSize)
{
   uint8_t  *pStart;
   uint8_t  *pEnd;
   uint32_t  dSize;
   uint32_t  dFree;
   
   pStart = pTCB->pStackStart;
   pEnd   = (uint8_t*)pStart + pTCB->wStackSize;
   dSize  = (uint32_t)pEnd - (uint32_t)pStart;
   dFree  = GetStackFreeCount(pStart, pEnd);
   
   *pSize = dSize;

   return(dFree);
} /* OSGetStackInfo */

/*************************************************************************/
/*  OSStatEnable                                                         */
/*                                                                       */
/*  Enable the statistic functionality.                                  */
/*                                                                       */
/*  In    : none                                                         */
/*  Out   : none                                                         */
/*  Return: none                                                         */
/*************************************************************************/
void OSStatEnable (void)
{
#if (_PERF_OPT_NO_STATISTIC >= 1)
   /*
    * Prevent compiler and lint warnings.
    */
   uint32_t dDummy = (uint32_t)StatUpdateTime + (uint32_t)StatTask;
   (void)dDummy;
   (void)StatStack;                
   (void)dStatHirResPeriod;
#else
   /* Create the Statistic task */
   OSTaskCreate(&TCBStat, StatTask, NULL, STAT_TASK_PRIO,
                StatStack, sizeof(StatStack), "[stat]");
#endif                
} /* OSStatEnable */

/*************************************************************************/
/*  OSStatGetCPULoad                                                     */
/*                                                                       */
/*  Return the CPU load in percent.                                      */
/*                                                                       */
/*  In    : none                                                         */
/*  Out   : none                                                         */
/*  Return: none                                                         */
/*************************************************************************/
uint8_t OSStatGetCPULoad (void)
{
   return(bStatCPULoad);
} /* OSStatGetCPULoad */

/*************************************************************************/
/*  OSStatGetTotalTaskTime                                               */
/*                                                                       */
/*  Return the total task time, of all tasks, from the last second.      */
/*                                                                       */
/*  In    : none                                                         */
/*  Out   : none                                                         */
/*  Return: none                                                         */
/*************************************************************************/
uint32_t OSStatGetTotalTaskTime (void)
{
   return(dStatTotalTaskTime);
} /* OSStatGetTotalTaskTime */

/*************************************************************************/
/*  OSTaskGetPriority                                                    */
/*                                                                       */
/*  Get the priority of the actual running task.                         */
/*                                                                       */
/*  In    : none                                                         */
/*  Out   : none                                                         */
/*  Return: The task priority                                            */
/*************************************************************************/
int OSTaskGetPriority (void)
{
   return(RunningTask->nPrio);
} /* OSTaskGetPriority */

/*************************************************************************/
/*  OSTaskChangePriority                                                 */
/*                                                                       */
/*  Change the priority of the actual running task.                      */
/*                                                                       */
/*  In    : Prio                                                         */
/*  Out   : none                                                         */
/*  Return: The old priority                                             */
/*************************************************************************/
int OSTaskChangePriority (int nPrio)
{
   int nOldPrio = RunningTask->nPrio;
   
   RunningTask->nPrio = nPrio;
   
   return(nOldPrio);
} /* OSTaskChangePriority */

/*************************************************************************/
/*  OSTaskYield                                                          */
/*                                                                       */
/*  Give CPU time to the next task from the ReadyList.                   */
/*                                                                       */
/*  In    : none                                                         */
/*  Out   : none                                                         */
/*  Return: none                                                         */
/*************************************************************************/
void OSTaskYield (void)
{
   EnterCritical();
   
   /*
    * Add the actual running task to the ReadyList.
    */
   AddTaskToReadyList(RunningTask);    
   
   /*
    * Schedule the next task
    */
   TaskSchedule();
   
   ExitCritical();             
} /* OSTaskYield */

/*************************************************************************/
/*  OSTaskExit                                                           */
/*                                                                       */
/*  Terminate the actual running task.                                   */
/*                                                                       */
/*  In    : none                                                         */
/*  Out   : none                                                         */
/*  Return: none                                                         */
/*************************************************************************/
void OSTaskExit (void)
{
   EnterCritical();
   
   /* Remove task from the task list */
   TaskListRemove(RunningTask);
   
   /* 
    * Set state to "not in use".
    *
    *   -- Do not use the macro SET_TASK_STATE here --
    *
    * The macro can be "empty" in case of performance
    * optimisation and will not work.
    */   
   RunningTask->State = OS_TASK_STATE_NOT_IN_USE;
   
   /* Schedule to the next task which is ready */
   TaskScheduleExit();
   
   ExitCritical();      
} /* OSTaskExit */

/*************************************************************************/
/*  OSTaskTerminateRequest                                               */
/*                                                                       */
/*  Set the termination flag only.                                       */
/*                                                                       */
/*  In    : none                                                         */
/*  Out   : none                                                         */
/*  Return: none                                                         */
/*************************************************************************/
void OSTaskTerminateRequest (os_tcb_t *pTCB)
{
   EnterCritical();

   if (pTCB != NULL)
   {
      pTCB->bFlagTermRequest = 1;
   }
   
   ExitCritical();      
} /* OSTaskTerminateRequest */

/*************************************************************************/
/*  OSTaskShouldTerminate                                                */
/*                                                                       */
/*  Return the termination flag of the task.                             */
/*                                                                       */
/*  In    : none                                                         */
/*  Out   : none                                                         */
/*  Return: termination flag                                             */
/*************************************************************************/
uint8_t OSTaskShouldTerminate (void)
{
   return(RunningTask->bFlagTermRequest);
} /* OSTaskShouldTerminate */

/*************************************************************************/
/*  OSTaskGetList                                                        */
/*                                                                       */
/*  Return the first element of the task list.                           */
/*                                                                       */
/*  In    : none                                                         */
/*  Out   : none                                                         */
/*  Return: first element of the task list                               */
/*************************************************************************/
os_tcb_t *OSTaskGetList (void)
{
   return(pTaskList);
} /* OSTaskGetList */

/*************************************************************************/
/*  OSTimerCallback                                                      */
/*                                                                       */
/*  This function will be called from the Systick interrupt.             */
/*                                                                       */
/*  Note: Interrupts are disabled and must be disabled.                  */
/*                                                                       */
/*  In    : none                                                         */
/*  Out   : none                                                         */
/*  Return: none                                                         */
/*************************************************************************/
void OSTimerCallback (void)
{
   static uint32_t dSecCnt = 0;
   
   /* 
    * System ticker 
    */
   dSystemTick++;

   /* 
    * Second ticker
    */
   dSecCnt++;
   if (OS_TICKS_PER_SECOND == dSecCnt)
   {
      dSecCnt = 0;
      dSecTick++;
      dUnixtime++;
   }

   /* 
    * Go through the WaitList and decrease the WaitCount. 
    */
   HandleWaitList();
   
   
   /*
    * Check user timer
    */
   if (UserTimer != NULL)
   {
      if (UserTimerTimeout > 0)
      {
         UserTimerTimeout--;
      }
      else
      {
         UserTimer();
         UserTimer = NULL;
      }
   }       
   
} /* OSTimerCallback */

/*************************************************************************/
/*  OSTimerSet                                                           */
/*                                                                       */
/*  Set a user timer and timeout.                                        */
/*                                                                       */
/*  In    : none                                                         */
/*  Out   : none                                                         */
/*  Return: none                                                         */
/*************************************************************************/
void OSTimerSet (OS_TIMER_FUNC Timer, uint32_t dTimeoutMs)
{
   uint32_t dTicks;
   
   dTicks = OS_MS_2_TICKS(dTimeoutMs);

   EnterCritical();
   
   UserTimer        = Timer;
   UserTimerTimeout = dTicks;

   ExitCritical();
   
} /* OSTimerSet */

/*************************************************************************/
/*  OSTimeGet                                                            */
/*                                                                       */
/*  Return the time from the system ticker.                              */
/*                                                                       */
/*  In    : none                                                         */
/*  Out   : none                                                         */
/*  Return: Systick                                                      */
/*************************************************************************/
uint32_t OSTimeGet (void)
{
   return(dSystemTick);
} /* OSTimeGet */

/*************************************************************************/
/*  OSTimeGetSeconds                                                     */
/*                                                                       */
/*  Return the time from the second ticker.                              */
/*                                                                       */
/*  Out   : none                                                         */
/*  Return: dSecTick                                                     */
/*************************************************************************/
uint32_t OSTimeGetSeconds (void)
{
   return(dSecTick);
} /* OSTimeGetSeconds */

/*************************************************************************/
/*  OSTimeDly                                                            */
/*                                                                       */
/*  Causes the current task to wait for a specified interval or, if the  */
/*  specified interval is zero, to give up the CPU for another task      */
/*  (with higher or same priority).                                      */
/*                                                                       */
/*  Note: Must not used from inside an interrupt.                        */
/*                                                                       */
/*  In    : TimeoutMs                                                    */
/*  Out   : none                                                         */
/*  Return: none                                                         */
/*************************************************************************/
void OSTimeDly (uint32_t dTimeoutMs)
{
   uint32_t dTicks;
   
   dTicks = OS_MS_2_TICKS(dTimeoutMs);
   
   if (0 == dTicks)
   {
      OSTaskYield();
   }
   else
   {
      EnterCritical();
      RunningTask->dTimeoutTicks = dTicks;
      AddTaskToWaitList(RunningTask);  /* Add the task to the WaitList */
      TaskSchedule();                  /* Schedule the next task */
      ExitCritical();
   }
   
} /* OSTimeDly */

/*************************************************************************/
/*  OSTimeDlyUntil                                                       */
/*                                                                       */
/*  Causes the current task to wait for a specified time.                */
/*                                                                       */
/*  Note: Must not used from inside an interrupt.                        */
/*                                                                       */
/*  In    : dTicks                                                       */
/*  Out   : none                                                         */
/*  Return: none                                                         */
/*************************************************************************/
void OSTimeDlyUntil (uint32_t dTicks)
{
   int64_t WaitTime;
   
   /* Calculate delta */
   WaitTime = dTicks - OSTimeGet(); 
   
   if (WaitTime <= 0)
   {
      OSTaskYield();
   }
   else
   {
      EnterCritical();
      RunningTask->dTimeoutTicks = (uint32_t)WaitTime;
      AddTaskToWaitList(RunningTask);  /* Add the task to the WaitList */
      TaskSchedule();                  /* Schedule the next task */
      ExitCritical();
   }
   
} /* OSTimeDlyUntil */

/*************************************************************************/
/*  OSTimeWait                                                           */
/*                                                                       */
/*  Loop for a specified number of milliseconds,                         */
/*  will not release the CPU.                                            */
/*                                                                       */
/*  In    : TimeoutMs                                                    */
/*  Out   : none                                                         */
/*  Return: none                                                         */
/*************************************************************************/
void OSTimeWait (uint32_t dTimeoutMs)
{
   uint32_t dTicks;
   uint32_t dTimeEnd;
   
   dTicks = OS_MS_2_TICKS(dTimeoutMs);
   
   dTimeEnd = dSystemTick + dTicks;
   
   while (dSystemTick <= dTimeEnd)
   {
      __asm__ volatile ("nop");
   }
      
} /* OSTimeWait */

/*************************************************************************/
/*  OSUnixtimeGet                                                        */
/*                                                                       */
/*  Return the unixtime.                                                 */
/*                                                                       */
/*  In    : none                                                         */
/*  Out   : none                                                         */
/*  Return: Unixtime                                                     */
/*************************************************************************/
uint32_t OSUnixtimeGet (void)
{
   return(dUnixtime);
} /* OSUnixtimeGet */

/*************************************************************************/
/*  OSUnixtimeSet                                                        */
/*                                                                       */
/*  Set a new unixtime.                                                  */
/*                                                                       */
/*  In    : dTime                                                        */
/*  Out   : none                                                         */
/*  Return: none                                                         */
/*************************************************************************/
void OSUnixtimeSet (uint32_t dTime)
{
   dUnixtime = dTime;
} /* OSUnixtimeSet */

/*************************************************************************/
/*  OSTimezoneIDSet                                                      */
/*                                                                       */
/*  Set the timezone ID.                                                 */
/*                                                                       */
/*  In    : wID                                                          */
/*  Out   : none                                                         */
/*  Return: none                                                         */
/*************************************************************************/
void OSTimezoneIDSet (int16_t wID)
{
   wTimezoneID = wID;
} /* OSTimezoneIDSet */

/*************************************************************************/
/*  OSTimezoneIDGet                                                      */
/*                                                                       */
/*  Return the timezone ID.                                              */
/*                                                                       */
/*  In    : none                                                         */
/*  Out   : none                                                         */
/*  Return: Timezone                                                     */
/*************************************************************************/
int16_t OSTimezoneIDGet (void)
{
   return(wTimezoneID);
} /* OSTimezoneIDGet */

/*************************************************************************/
/*  OSTimezoneMinSet                                                     */
/*                                                                       */
/*  Set the timezone offset to UTC in minutes.                           */
/*                                                                       */
/*  In    : wOffset                                                       */
/*  Out   : none                                                         */
/*  Return: none                                                         */
/*************************************************************************/
void OSTimezoneMinSet (int16_t wOffset)
{
   wTimezoneOffsetMin = wOffset;
   dTimezoneOffsetSec = wOffset * 60;
} /* OSTimezoneMinSet */

/*************************************************************************/
/*  OSTimezoneMinGet                                                     */
/*                                                                       */
/*  Return the timezone offset in minutes to UTC.                        */
/*                                                                       */
/*  In    : none                                                         */
/*  Out   : none                                                         */
/*  Return: Timezone                                                     */
/*************************************************************************/
int16_t OSTimezoneMinGet (void)
{
   return(wTimezoneOffsetMin);
} /* OSTimezoneMinGet */

/*************************************************************************/
/*  OSTimezoneSecGet                                                     */
/*                                                                       */
/*  Return the timezone offset in seconds to UTC.                        */
/*                                                                       */
/*  In    : none                                                         */
/*  Out   : none                                                         */
/*  Return: Timezone                                                     */
/*************************************************************************/
int32_t OSTimezoneSecGet (void)
{
   return(dTimezoneOffsetSec);
} /* OSTimezoneSecGet */

/*************************************************************************/
/*  OSTimezoneDstSet                                                     */
/*                                                                       */
/*  Set the daylight saving time offset to UTC in hours.                 */
/*                                                                       */
/*  In    : bOffset                                                      */
/*  Out   : none                                                         */
/*  Return: none                                                         */
/*************************************************************************/
void OSTimezoneDstSet (int8_t bOffset)
{
   dTimezoneDstOffsetSec = bOffset * 60 * 60;
} /* OSTimezoneDstSet */

/*************************************************************************/
/*  OSTimezoneDstSecGet                                                  */
/*                                                                       */
/*  Return the daylight saving time offset in seconds to UTC.            */
/*                                                                       */
/*  In    : none                                                         */
/*  Out   : none                                                         */
/*  Return: Timezone                                                     */
/*************************************************************************/
int32_t OSTimezoneDstSecGet (void)
{
   return(dTimezoneDstOffsetSec);
} /* OSTimezoneDstSecGet */

/*************************************************************************/
/*  OSSemaCreate                                                         */
/*                                                                       */
/*  Create a new semaphore.                                              */
/*                                                                       */
/*  In    : pSema, dCounterStart, dCounterMax                            */
/*  Out   : none                                                         */
/*  Return: none                                                         */
/*************************************************************************/
void OSSemaCreate (os_sema_t *pSema, int32_t nCounterStart, int32_t nCounterMax)
{
   memset(pSema, 0x00, sizeof(os_sema_t));

   pSema->nCounter    = nCounterStart;
   pSema->nCounterMax = nCounterMax;   
} /* OSSemaCreate */

/*************************************************************************/
/*  OSSemaReset                                                          */
/*                                                                       */
/*  Reset a semaphore. All waiting tasks will be released.               */
/*                                                                       */
/*  In    : pSema, nCounterStart                                         */
/*  Out   : none                                                         */
/*  Return: none                                                         */
/*************************************************************************/
void OSSemaReset (os_sema_t *pSema, int32_t nCounterStart)
{
   os_tcb_t *pTask;

   EnterCritical();

   /* Set new counter */   
   pSema->nCounter = nCounterStart;
   
   /* Check if fifo is not empty  */
   if (0 == IsFifoEmpty(&pSema->Fifo))
   {
      /* Fifo not empty, remove task from the semaphore fifo */
      pTask = TCBFifoRemove(&pSema->Fifo);
      while (pTask != NULL)
      {
         /* Remove semaphore info */
         pTask->pSemaWait = NULL;
      
         /* Removed the task from the WaitList too. */
         WaitListRemove(pTask);
         pTask->dTimeoutTicks = 0;

         /* Return code for OSSemaWait */
         pTask->nReturnCode = OS_RC_RESETED;

         /* Make the task which is waiting on the semaphore ready to run */
         AddTaskToReadyList(pTask);

         /* Remove the next task if exist */         
         if (0 == IsFifoEmpty(&pSema->Fifo))
         {
            /* Fifo not empty, remove task from the semaphore fifo */
            pTask = TCBFifoRemove(&pSema->Fifo);
         }
         else
         {
            break;
         }   
      }         

      /* Add the task which reset the semaphore to the ReadyList */ 
      AddTaskToReadyList(RunningTask);   

      TaskSchedule(); /* Schedule the next task */
   }

   ExitCritical();   
} /* OSSemaReset */

/*************************************************************************/
/*  OSSemaDelete                                                         */
/*                                                                       */
/*  Delete a semaphore. All waiting tasks will be released.              */
/*                                                                       */
/*  In    : pSema                                                        */
/*  Out   : none                                                         */
/*  Return: none                                                         */
/*************************************************************************/
void OSSemaDelete (os_sema_t *pSema)
{
   os_tcb_t *pTask;

   EnterCritical();

   /* Check if fifo is not empty  */
   if (0 == IsFifoEmpty(&pSema->Fifo))
   {
      /* Fifo not empty, remove task from the semaphore fifo */
      pTask = TCBFifoRemove(&pSema->Fifo);
      while (pTask != NULL)
      {
         /* Remove semaphore info */
         pTask->pSemaWait = NULL;
      
         /* Removed task from the WaitList too. */
         WaitListRemove(pTask);
         pTask->dTimeoutTicks = 0;

         /* Return code for OSSemaWait */
         pTask->nReturnCode = OS_RC_DELETED;

         /* Make the task which is waiting on the semaphore ready to run */
         AddTaskToReadyList(pTask);

         /* Remove the next task if exist */         
         if (0 == IsFifoEmpty(&pSema->Fifo))
         {
            /* Fifo not empty, remove task from the semaphore fifo */
            pTask = TCBFifoRemove(&pSema->Fifo);
         }
         else
         {
            break;
         }   
      }
   }
   
   ExitCritical();   
} /* OSSemaDelete */

/*************************************************************************/
/*  _OSSemaSignalFromInt                                                 */
/*                                                                       */
/*  Signal a semaphore from interrupt.                                   */
/*                                                                       */
/*  Must be called with interrupts disabled.                             */
/*                                                                       */
/*  Note: Interrupts are disabled and must be disabled.                  */
/*                                                                       */
/*  In    : pSema                                                        */
/*  Out   : none                                                         */
/*  Return: 1 = must schedule / 0 = schedule not needed / error cause    */
/*************************************************************************/
static __inline__ int _OSSemaSignalFromInt (os_sema_t *pSema)
{
   int        rc = 0;
   os_tcb_t *pTask;
   
   /* Check if fifo is not empty  */
   if (0 == IsFifoEmpty(&pSema->Fifo))
   {
      /* Fifo not empty, remove task from the semaphore fifo */
      pTask = TCBFifoRemove(&pSema->Fifo);

      /* Remove semaphore info */
      pTask->pSemaWait = NULL;
      
      /* Removed the task from the WaitList too. */
      WaitListRemove(pTask);
      pTask->dTimeoutTicks = 0;

      /* Return code for OSSemaWait */
      pTask->nReturnCode = OS_RC_OK;

      /* Make the task which is waiting on the semaphore ready to run */
      AddTaskToReadyList(pTask);
      
      rc = 1;
   }
   else
   {
      /* Fifo was empty, check for max value */
      if (pSema->nCounter < pSema->nCounterMax)
      {
         pSema->nCounter++;
      }
      else
      {
         rc = OS_RC_NO_SPACE;
      }   
   } /* end if (0 == IsFifoEmpty(&pSema->Fifo)) */
   
   return(rc);
} /* _OSSemaSignalFromInt */

/*************************************************************************/
/*  OSSemaSignalFromInt                                                  */
/*                                                                       */
/*  Signal a semaphore from interrupt.                                   */
/*                                                                       */
/*  Must be called with interrupts disabled.                             */
/*                                                                       */
/*  Note: Interrupts are disabled and must be disabled.                  */
/*                                                                       */
/*  In    : pSema                                                        */
/*  Out   : none                                                         */
/*  Return: 1 = must schedule / 0 = schedule not needed / error cause    */
/*************************************************************************/
int OSSemaSignalFromInt (os_sema_t *pSema)
{
   return(_OSSemaSignalFromInt(pSema));
} /* OSSemaSignalFromInt */

/*************************************************************************/
/*  OSSemaSignal                                                         */
/*                                                                       */
/*  Signal a semaphore.                                                  */
/*                                                                       */
/*  Note: Must not used from inside an interrupt.                        */
/*                                                                       */
/*  In    : pSema                                                        */
/*  Out   : none                                                         */
/*  Return: OS_RC_OK / error cause                                       */
/*************************************************************************/
int OSSemaSignal (os_sema_t *pSema)
{
   int rc;
   
   EnterCritical();
   
   rc = _OSSemaSignalFromInt(pSema);

   /* Check if we must schedule */
   if (1 == rc)
   {   
      /* Add the task which signal the semaphore to the ReadyList */ 
      AddTaskToReadyList(RunningTask);   

      TaskSchedule(); /* Schedule the next task */
      
      rc = OS_RC_OK;         
   }      
   
   ExitCritical();
   
   return(rc);
} /* OSSemaSignal */

/*************************************************************************/
/*  OSSemaSignalNoSched                                                  */
/*                                                                       */
/*  Signal a semaphore but without scheduling.                           */
/*                                                                       */
/*  Note: Must not used from inside an interrupt.                        */
/*                                                                       */
/*  In    : pSema                                                        */
/*  Out   : none                                                         */
/*  Return: OS_RC_OK / error cause                                       */
/*************************************************************************/
int OSSemaSignalNoSched (os_sema_t *pSema)
{
   int rc;
   
   EnterCritical();
   
   rc = OSSemaSignalFromInt(pSema);
   
   ExitCritical();
   
   return(rc);
} /* OSSemaSignalNoSched */

/*************************************************************************/
/*  OSSemaWait                                                           */
/*                                                                       */
/*  Blocks the task while waiting for the semaphore, with timeout.       */
/*                                                                       */
/*  In case the timeout value (dTimeoutMs) is 0, OSSemaWait will use     */
/*  polling to acquire the resource. If the resource counter (dCounter)  */
/*  is 0, the error cause is OS_RC_TIMEOUT.                              */
/*                                                                       */
/*  With a timeout value (dTimeoutMs) of OS_WAIT_INFINITE the waiting    */
/*  time is not ended until the semaphore is signaled.                   */
/*                                                                       */
/*  Note: Must not used from inside an interrupt.                        */
/*                                                                       */
/*  In    : pSema, dTimeoutMs                                            */
/*  Out   : none                                                         */
/*  Return: OS_RC_OK / error casue                                       */
/*************************************************************************/
int OSSemaWait (os_sema_t *pSema, uint32_t dTimeoutMs)
{
   int      rc = OS_RC_OK;
   uint32_t dTicks;
   
   EnterCritical();
   
   /* Check if semaphore is available */
   if (pSema->nCounter > 0)
   {
      pSema->nCounter--;
   }   
   else
   {
      dTicks = ((dTimeoutMs != OS_WAIT_INFINITE) ? OS_MS_2_TICKS(dTimeoutMs) : OS_WAIT_INFINITE);
      if (dTicks != 0)
      {
         /* Not available, prepare the task for waiting */
         RunningTask->dTimeoutTicks = dTicks;
         RunningTask->pSemaWait     = pSema;
         RunningTask->nReturnCode   = OS_RC_ERROR;
         SET_TASK_STATE(RunningTask, OS_TASK_STATE_WAITING);

         /* Add the task to the fifo of the semaphore */
         TCBFifoAddPrio(&pSema->Fifo, RunningTask);
      
         /*  Add the task to the WaitList too */
         AddTaskToWaitList(RunningTask);      

         TaskSchedule(); /* Schedule the next task */
      
         /* This is the return code of the function */
         rc = RunningTask->nReturnCode;
      }
      else
      {
         /* Polling, no resource available */
         rc = OS_RC_TIMEOUT; 
      }         
   }
   
   ExitCritical();
   
   return(rc);
} /* OSSemaWait */

/*************************************************************************/
/*  OSSemaBroadcastAsync                            (needed for NutNET)  */
/*                                                                       */
/*  Wake up all tasks waiting on this semaphore. But even if the         */
/*  priority of any woken task is higher than the current tasks's        */
/*  priority, the current one continues running.                         */
/*                                                                       */
/*  Note: No scheduling here, must not used from inside an interrupt.    */
/*                                                                       */
/*  In    : pSema                                                        */
/*  Out   : none                                                         */
/*  Return: The number of tasks woken up / error cause                   */
/*************************************************************************/
int OSSemaBroadcastAsync (os_sema_t *pSema)
{
   int         rc = 0;
   os_tcb_t *pTask;
   
   EnterCritical();

   /* Check if fifo is not empty  */
   if (0 == IsFifoEmpty(&pSema->Fifo))
   {
      /* Fifo not empty, remove task from the semaphore fifo */
      pTask = TCBFifoRemove(&pSema->Fifo);
      while (pTask != NULL) 
      {
         /* Remove semaphore info */
         pTask->pSemaWait = NULL;
      
         /* Removed the task from the WaitList */
         WaitListRemove(pTask);
         pTask->dTimeoutTicks = 0;

         /* Return code for OSSemaWait */
         pTask->nReturnCode = OS_RC_OK;

         /* Make the task which is waiting on the semaphore ready to run */
         AddTaskToReadyList(pTask);
         
         rc++;

         /* Get the next task from the fifo */         
         if (0 == IsFifoEmpty(&pSema->Fifo))
         {
            /* Fifo not empty, remove task from the semaphore fifo */
            pTask = TCBFifoRemove(&pSema->Fifo);
         }          
         else
         {
            break;
         }   
      }          
   }
   else
   {
      /* Fifo was empty, check for max value */
      if (pSema->nCounter < pSema->nCounterMax)
      {
         pSema->nCounter++;
      }
      else
      {
         rc = OS_RC_NO_SPACE;
      }   
   } /* end if (0 == IsFifoEmpty(&pSema->Fifo)) */
   
   ExitCritical();
   
   return(rc);
} /* OSSemaBroadcastAsync */

/*************************************************************************/
/*  OSSemaBroadcast                                 (needed for NutNET)  */
/*                                                                       */
/*  Wake up all tasks waiting on this semaphore. If the priority of any  */
/*  waiting tasks is higher or equal than the current task's priority,   */
/*  then the current task is stopped and CPU control is passed to the    */
/*  woken up task with the highest priority.                             */
/*                                                                       */
/*  Note: Must not used from inside an interrupt.                        */
/*                                                                       */
/*  Out   : none                                                         */
/*  Return: The number of tasks woken up / error cause                   */
/*************************************************************************/
int OSSemaBroadcast (os_sema_t *pSema)
{
   int rc;
   
   rc = OSSemaBroadcastAsync(pSema);
   
   /*
    * If any tasks with higher or equal priority is
    * ready to run, switch the context.
    */
   OSTaskYield();
   
   return(rc);
} /* OSSemaBroadcast */  

/*************************************************************************/
/*  OSMutexCreate                                                        */
/*                                                                       */
/*  Create a new mutex.                                                  */
/*                                                                       */
/*  In    : pMutex                                                       */
/*  Out   : none                                                         */
/*  Return: none                                                         */
/*************************************************************************/
void OSMutexCreate (os_mutex_t *pMutex)
{
   memset(pMutex, 0x00, sizeof(os_mutex_t));
} /* OSMutexCreate */

/*************************************************************************/
/*  _OSMutexSignalFromInt                                                */
/*                                                                       */
/*  Signal a mutex from interrupt.                                       */
/*                                                                       */
/*  Must be called with interrupts disabled.                             */
/*                                                                       */
/*  Note: Interrupts are disabled and must be disabled.                  */
/*                                                                       */
/*  In    : pMutex                                                       */
/*  Out   : none                                                         */
/*  Return: 1 = must schedule / 0 = schedule not needed / error cause    */
/*************************************************************************/
static __inline__ int _OSMutexSignalFromInt (os_mutex_t *pMutex)
{
   int        rc = 0;
   os_tcb_t *pTask;
   
   /* Check if fifo is not empty  */
   if (0 == IsFifoEmpty(&pMutex->Fifo))
   {
      if (pMutex->nCounter > 0)
      {
         pMutex->nCounter--;
      }
      
      if (0 == pMutex->nCounter)
      {  
         pMutex->Owner = NULL;
          
         /* Fifo not empty, remove task from the mutex fifo */
         pTask = TCBFifoRemove(&pMutex->Fifo);

         /* Remove semaphore info */
         pTask->pMutexWait = NULL;
      
         /* Removed the task from the WaitList too. */
         WaitListRemove(pTask);
         pTask->dTimeoutTicks = 0;

         /* Return code for OSSemaWait */
         pTask->nReturnCode = OS_RC_OK;

         /* Make the task which is waiting on the semaphore ready to run */
         AddTaskToReadyList(pTask);
      }         
      
      rc = 1;
   }
   else
   {
      /* Fifo was empty, check for max value */
      if (pMutex->nCounter > 0)
      {
         pMutex->nCounter--;
         if (0 == pMutex->nCounter)
         {
            pMutex->Owner = NULL;
         }
      }
      else
      {
         rc = OS_RC_ERROR;
      }   
   } /* end if (0 == IsFifoEmpty(&pSema->Fifo)) */
   
   return(rc);
} /* _OSMutexSignalFromInt */

/*************************************************************************/
/*  OSMutexSignalFromInt                                                 */
/*                                                                       */
/*  Signal a mutex from interrupt.                                       */
/*                                                                       */
/*  Must be called with interrupts disabled.                             */
/*                                                                       */
/*  Note: Interrupts are disabled and must be disabled.                  */
/*                                                                       */
/*  In    : pMutex                                                       */
/*  Out   : none                                                         */
/*  Return: 1 = must schedule / 0 = schedule not needed / error cause    */
/*************************************************************************/
int OSMutexSignalFromInt (os_mutex_t *pMutex)
{
   return(_OSMutexSignalFromInt(pMutex));
} /* OSMutexSignalFromInt */

/*************************************************************************/
/*  OSMutexSignal                                                        */
/*                                                                       */
/*  Signal a mutex.                                                      */
/*                                                                       */
/*  Note: Must not used from inside an interrupt.                        */
/*                                                                       */
/*  In    : pSema                                                        */
/*  Out   : none                                                         */
/*  Return: OS_RC_OK / error cause                                       */
/*************************************************************************/
int OSMutexSignal (os_mutex_t *pMutex)
{
   int rc;
   
   EnterCritical();
   
   rc = _OSMutexSignalFromInt(pMutex);

   /* Check if we must schedule */
   if (1 == rc)
   {   
      /* Add the task which signal the mutex to the ReadyList */ 
      AddTaskToReadyList(RunningTask);   

      TaskSchedule(); /* Schedule the next task */
      
      rc = OS_RC_OK;         
   }      
   
   ExitCritical();
   
   return(rc);
} /* OSMutexSignal */

/*************************************************************************/
/*  OSMutexWait                                                          */
/*                                                                       */
/*  Blocks the task while waiting for the mutex, with timeout.           */
/*                                                                       */
/*  In case the timeout value (dTimeoutMs) is 0, OSSemaWait will use     */
/*  polling to acquire the resource. If the resource counter (dCounter)  */
/*  is 0, the error cause is OS_RC_TIMEOUT.                              */
/*                                                                       */
/*  With a timeout value (dTimeoutMs) of OS_WAIT_INFINITE the waiting    */
/*  time is not ended until the semaphore is signaled.                   */
/*                                                                       */
/*  Note: Must not used from inside an interrupt.                        */
/*                                                                       */
/*  In    : pSema, dTimeoutMs                                            */
/*  Out   : none                                                         */
/*  Return: OS_RC_OK / error casue                                       */
/*************************************************************************/
int OSMutexWait (os_mutex_t *pMutex, uint32_t dTimeoutMs)
{
   int      rc = OS_RC_OK;
   uint32_t dTicks;
   
   EnterCritical();
   
   /* Check if mutex is available */
   if (NULL == pMutex->Owner)
   {
      pMutex->Owner = RunningTask; 
      pMutex->nCounter++;
   }   
   else if (pMutex->Owner == RunningTask)
   {
      pMutex->nCounter++;
   }
   else
   {
      dTicks = ((dTimeoutMs != OS_WAIT_INFINITE) ? OS_MS_2_TICKS(dTimeoutMs) : OS_WAIT_INFINITE);
      if (dTicks != 0)
      {
         /* Not available, prepare the task for waiting */
         RunningTask->dTimeoutTicks = dTicks;
         RunningTask->pMutexWait    = pMutex;
         RunningTask->nReturnCode   = OS_RC_ERROR;
         SET_TASK_STATE(RunningTask, OS_TASK_STATE_WAITING);

         /* Add the task to the fifo of the mutex */
         TCBFifoAddPrio(&pMutex->Fifo, RunningTask);
      
         /*  Add the task to the WaitList too */
         AddTaskToWaitList(RunningTask);      

         TaskSchedule(); /* Schedule the next task */
      
         /* This is the return code of the function */
         rc = RunningTask->nReturnCode;
      }
      else
      {
         /* Polling, no resource available */
         rc = OS_RC_TIMEOUT; 
      }         
   }
   
   ExitCritical();
   
   return(rc);
} /* OSMutexWait */

/*************************************************************************/
/*  OSEventCreate                                                        */
/*                                                                       */
/*  Create a new event.                                                  */
/*                                                                       */
/*  In    : pEvent                                                       */
/*  Out   : none                                                         */
/*  Return: none                                                         */
/*************************************************************************/
void OSEventCreate (os_event_t *pEvent)
{
   memset(pEvent, 0x00, sizeof(os_event_t));

   pEvent->dPattern = 0;
} /* OSEventCreate */

/*************************************************************************/
/*  OSEventDelete                                                        */
/*                                                                       */
/*  Delete an event.                                                     */
/*                                                                       */
/*  In    : pEvent                                                       */
/*  Out   : none                                                         */
/*  Return: none                                                         */
/*************************************************************************/
void OSEventDelete (os_event_t *pEvent)
{
   os_tcb_t *pTask;

   EnterCritical();

   /* Check if fifo is not empty  */
   if (0 == IsFifoEmpty(&pEvent->Fifo))
   {
      /* Fifo not empty, remove task from the event fifo */
      pTask = TCBFifoRemove(&pEvent->Fifo);
      while (pTask != NULL)
      {
         /* Remove event info */
         pTask->pEventWait = NULL;
         pTask->dEventWaitPattern = 0;
      
         /* Removed the task from the WaitList too. */
         WaitListRemove(pTask);
         pTask->dTimeoutTicks = 0;

         /* Return code for OSEventWait */
         pTask->nReturnCode = OS_RC_DELETED;

         /* Make the task which is waiting on the event ready to run */
         AddTaskToReadyList(pTask);

         /* Remove the next task if exist */         
         if (0 == IsFifoEmpty(&pEvent->Fifo))
         {
            /* Fifo not empty, remove task from the event fifo */
            pTask = TCBFifoRemove(&pEvent->Fifo);
         }
         else
         {
            break;
         }   
      }
   }
   
   ExitCritical();
} /* OSEventDelete */

/*************************************************************************/
/*  OSEventSetFromInt                                                    */
/*                                                                       */
/*  Signal an event from interrupt.                                      */
/*                                                                       */
/*  Must be called with interrupts disabled.                             */
/*                                                                       */
/*  Note: Interrupts are disabled and must be disabled.                  */
/*                                                                       */
/*  In    : pEvent, dPattern                                             */
/*  Out   : none                                                         */
/*  Return: none                                                         */
/*************************************************************************/
void OSEventSetFromInt (os_event_t *pEvent, uint32_t dPattern)
{
   int        rc = 0;
   int        GoReady;
   os_tcb_t *pTask;
   os_tcb_t *pPrev;

   pEvent->dPattern |= dPattern;
   
   /* Check if fifo is NOT empty  */
   if (0 == IsFifoEmpty(&pEvent->Fifo))
   {
      /* Fifo is not empty */
      pTask = pEvent->Fifo.pOut;

      while (pTask != NULL)
      {
         /* 
          * Save the prev pointer of the task.
          * pTask->pPrev could not be used later, because it wil be changed
          * by the TCBFifoRemoveMiddle before we can use the Prev for the loop.
          */
         pPrev = pTask->pPrev;
      
         /*
          * Check if the task must wait or not
          */
         if (OS_EVENT_MODE_OR == pTask->EventWaitMode)
         {
            GoReady = ((pEvent->dPattern & pTask->dEventWaitPattern) != 0) ? 1 : 0;   
         }
         else
         {
            GoReady = ((pEvent->dPattern & pTask->dEventWaitPattern) == pTask->dEventWaitPattern) ? 1 : 0; 
         }
   
         if (GoReady)
         {
            pTask->dEventWaitPattern = pEvent->dPattern;
            pTask->nReturnCode       = OS_RC_OK;
         
            TCBFifoRemoveMiddle(&pEvent->Fifo, pTask);
         
            /* Remove event info */
            pTask->pEventWait = NULL;
                  
            /* Removed the task from the WaitList too. */
            WaitListRemove(pTask);
            pTask->dTimeoutTicks = 0;
         
            AddTaskToReadyList(pTask);
         
            rc++;
         }
   
         pTask = pPrev;
      }
      
      if (rc != 0)
      {
         /* Event occurred, the pattern can be cleared */
         pEvent->dPattern = 0;
      }
   }

} /* OSEventSetFromInt */

/*************************************************************************/
/*  OSEventSet                                                           */
/*                                                                       */
/*  Signal an event.                                                     */
/*                                                                       */
/*  Note: Must not used from inside an interrupt.                        */
/*                                                                       */
/*  In    : pEvent, dPattern                                             */
/*  Out   : none                                                         */
/*  Return: none                                                         */
/*************************************************************************/
void OSEventSet (os_event_t *pEvent, uint32_t dPattern)
{
   EnterCritical();
   
   OSEventSetFromInt(pEvent, dPattern);
   
   /* Add the task which signal the event to the ReadyList */ 
   AddTaskToReadyList(RunningTask);   

   TaskSchedule(); /* Schedule the next task */
   
   ExitCritical();
} /* OSEventSet */

/*************************************************************************/
/*  OSEventSetNoSched                                                    */
/*                                                                       */
/*  Signal an event but without scheduling.                              */
/*                                                                       */
/*  Note: Must not used from inside an interrupt.                        */
/*                                                                       */
/*  In    : pEvent, dPattern                                             */
/*  Out   : none                                                         */
/*  Return: none                                                         */
/*************************************************************************/
void OSEventSetNoSched (os_event_t *pEvent, uint32_t dPattern)
{
   EnterCritical();
   
   OSEventSetFromInt(pEvent, dPattern);
      
   ExitCritical();
} /* OSEventSetNoSched */

/*************************************************************************/
/*  OSEventWait                                                          */
/*                                                                       */
/*  Blocks the task while waiting for the event, with timeout.           */
/*                                                                       */
/*  With a timeout value (dTimeoutMs) of OS_WAIT_INFINITE the waiting    */
/*  time is not ended until the event is signaled.                       */
/*                                                                       */
/*  Note: Must not used from inside an interrupt.                        */
/*                                                                       */
/*  In    : pEvent, dWaitPattern, Mode, pPattern, dTimeoutMs             */
/*  Out   : none                                                         */
/*  Return: OS_EVENT_RC_OK / error casue                                 */
/*************************************************************************/
int OSEventWait (os_event_t *pEvent, uint32_t dWaitPattern, 
                 os_event_mode_t Mode, uint32_t *pPattern, uint32_t dTimeoutMs)
{
   int       rc      = OS_RC_ERROR;
   int       GoReady = 0;
   uint32_t dTicks;   
   
   dTicks = ((dTimeoutMs != OS_WAIT_INFINITE) ? OS_MS_2_TICKS(dTimeoutMs) : OS_WAIT_INFINITE);
   
   EnterCritical();
   
   *pPattern = 0;
   
   /*
    * Check if the task must wait or not
    */
   if (OS_EVENT_MODE_OR == Mode)
   {
      GoReady = ((pEvent->dPattern & dWaitPattern) != 0) ? 1 : 0;   
   }
   else
   {
      GoReady = ((pEvent->dPattern & dWaitPattern) == dWaitPattern) ? 1 : 0; 
   }
   
   if (GoReady)
   {
      *pPattern = pEvent->dPattern;
      pEvent->dPattern = 0;
      
      rc = OS_RC_OK;
   }
   else
   {
      if (dTicks != 0)
      {
         /* Prepare the task for waiting */
         RunningTask->dTimeoutTicks     = dTicks;
         RunningTask->pEventWait        = pEvent;
         RunningTask->dEventWaitPattern = dWaitPattern;
         RunningTask->EventWaitMode     = Mode;
         RunningTask->nReturnCode       = OS_RC_ERROR;
         SET_TASK_STATE(RunningTask, OS_TASK_STATE_WAITING);
      
         /* Add the task to the fifo of the event */
         TCBFifoAddPrio(&pEvent->Fifo, RunningTask);
      
         /*  Add the task to the WaitList too */
         AddTaskToWaitList(RunningTask);      

         TaskSchedule(); /* Schedule the next task */

         /* This is the return code of the function */
         rc = RunningTask->nReturnCode;
         if (OS_RC_OK == rc)
         {
            *pPattern = RunningTask->dEventWaitPattern;
         }
      }
      else
      {
         /* Polling, no event available */
         rc = OS_RC_TIMEOUT; 
      }         
   }
      
   ExitCritical();
   
   return(rc);
} /* OSEventWait */

/*************************************************************************/
/*  OSMboxCreate                                                         */
/*                                                                       */
/*  Create a new mailbox.                                                */
/*                                                                       */
/*  In    : pMbox, pBuffer, nCounterMax                                  */
/*  Out   : none                                                         */
/*  Return: none                                                         */
/*************************************************************************/
void OSMboxCreate (os_mbox_t *pMbox, void **pBuffer, uint16_t wCounterMax)
{
   memset(pMbox, 0x00, sizeof(os_mbox_t));

   OSSemaCreate(&pMbox->UsedCntSema, 0,           wCounterMax); 
   OSSemaCreate(&pMbox->FreeCntSema, wCounterMax, wCounterMax); 

   pMbox->wCountMax = wCounterMax;
   pMbox->wCount    = 0;
   pMbox->wInIndex  = 0;
   pMbox->wOutIndex = 0;
   pMbox->pBuffer   = pBuffer;
} /* OSMboxCreate */

/*************************************************************************/
/*  OSMboxDelete                                                         */
/*                                                                       */
/*  Delete a mailbox. All waiting tasks will be released.                */
/*                                                                       */
/*  In    : pMbox                                                        */
/*  Out   : none                                                         */
/*  Return: none                                                         */
/*************************************************************************/
void OSMboxDelete (os_mbox_t *pMbox)
{
   OSSemaDelete(&pMbox->UsedCntSema);
   OSSemaDelete(&pMbox->FreeCntSema);
} /* OSMboxDelete */

/*************************************************************************/
/*  OSMboxPostFromInt                                                    */
/*                                                                       */
/*  Tries to post a message to the mailbox.                              */
/*                                                                       */
/*  Must be called with interrupts disabled.                             */
/*                                                                       */
/*  Note: Interrupts are disabled and must be disabled.                  */
/*                                                                       */
/*  In    : pMbox, pMsg                                                  */
/*  Out   : none                                                         */
/*  Return: OS_RC_OK / OS_RC_NO_SPACE                                    */
/*************************************************************************/
int OSMboxPostFromInt (os_mbox_t *pMbox, void *pMsg)
{
   int        rc = OS_RC_NO_SPACE;
   uint16_t wFreeCnt;

   wFreeCnt = pMbox->wCountMax - pMbox->wCount;
   if (wFreeCnt != 0)
   {
      /*
       * OSSemaWait cannot be used inside an interrupt.
       * Therefore decrement the semaphore counter direct.
       */
      pMbox->FreeCntSema.nCounter--;
   
      /* Increment counter, one more message available */
      pMbox->wCount++;
      
      /* Store message */
      pMbox->pBuffer[pMbox->wInIndex] = pMsg;
      
      /* Switch to next place */
      pMbox->wInIndex++;
      
      /* Check end of ring buffer */
      if (pMbox->wCountMax == pMbox->wInIndex)
      {
         pMbox->wInIndex = 0;
      }
   
      /* Signal message available */
      OSSemaSignalFromInt(&pMbox->UsedCntSema);
      
      rc = OS_RC_OK;
   }
   
   return(rc);
} /* OSMboxPostFromInt */

/*************************************************************************/
/*  OSMboxWait                                                           */
/*                                                                       */
/*  Blocks the task while waiting for the mailbox, with timeout.         */
/*                                                                       */
/*  In case the timeout value (dTimeoutMs) is 0, OSMboxWait will use     */
/*  polling to acquire the message. If the resource counter (dCounter)   */
/*  is 0, the error cause is OS_RC_TIMEOUT.                              */
/*                                                                       */
/*  With a timeout value (dTimeoutMs) of OS_WAIT_INFINITE the waiting    */
/*  time is not ended until the mailbox is signaled.                     */
/*                                                                       */
/*  Note: Must not used from inside an interrupt.                        */
/*                                                                       */
/*  In    : pMbox, pMsg, dTimeoutMs                                      */
/*  Out   : none                                                         */
/*  Return: OS_RC_OK / OS_RC_TIMEOUT                                     */
/*************************************************************************/
int OSMboxWait (os_mbox_t *pMbox, void **pMsg, uint32_t dTimeoutMs)
{
   int rc;
   
   if (0 == dTimeoutMs)
   {
      if (0 == pMbox->wCount)
      {
         return(OS_RC_TIMEOUT);
      }
   }
   
   rc = OSSemaWait(&pMbox->UsedCntSema, dTimeoutMs);
   if (OS_RC_OK == rc)
   {
      EnterCritical();

      /* Decrement counter, one message will be removed */
      pMbox->wCount--;

      /* Get message */      
      *pMsg = pMbox->pBuffer[pMbox->wOutIndex];

      /* Switch to next place */
      pMbox->wOutIndex++;

      /* Check end of ring buffer */
      if (pMbox->wCountMax == pMbox->wOutIndex)
      {
         pMbox->wOutIndex = 0;
      }

      OSSemaSignalFromInt(&pMbox->FreeCntSema);
      
      ExitCritical();
   }      
   
   return(rc);
} /* OSMboxWait*/

/*** EOF ***/
